#==============================================================================
#
#  $Id: mailbx.py,v 1.6 2001/03/13 02:28:03 mike Exp $
#
"""
   Module to manage a message database.
"""
#
#  Copyright (C) 1999 Michael A. Muller
#
#  Permission is granted to use, modify and redistribute this code,
#  providing that the following conditions are met:
#
#  1) This copyright/licensing notice must remain intact.
#  2) If the code is modified and redistributed, the modifications must 
#  include documentation indicating that the code has been modified.
#  3) The author(s) of this code must be indemnified and held harmless
#  against any damage resulting from the use of this code.
#
#  This code comes with ABSOLUTELY NO WARRANTEE, not even the implied 
#  warrantee of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
#
#  $Log: mailbx.py,v $
#  Revision 1.6  2001/03/13 02:28:03  mike
#  Added checkMailbox() method, which does a comparison of the index
#  vs. the actual contents of a database.
#
#  Revision 1.5  2000/05/11 00:59:19  mike
#  An even _safer_ wrapper around delete failures.
#
#  Revision 1.4  2000/04/27 23:17:43  mike
#  Added "special mailboxes", which are created if they do not exist; added
#  exception handling and logging around the deletion of a mail message, to try
#  to resolve mysterious exceptions.
#
#  Revision 1.3  2000/04/14 01:20:04  mike
#  Added PostOffice class and cleaned up mailbox so that it can deal with
#  pristine mailbox files.
#
#  Revision 1.2  2000/04/04 01:12:01  mike
#  Added delete message functionality.
#
#  Revision 1.1.1.1  2000/04/02 19:09:03  mike
#  Spug Mailer and libraries.
#
#
#==============================================================================

import shelve, os, re
from message import MessageInfo

class MailboxMessageInfo(MessageInfo):
   
   """
      A message info object with an associated reference to the @Mailbox that
      it is contained in.
   """      
   
   def __init__(self, sender, recipients, date, subject, unread, index,
                mailbox
                ):
      MessageInfo.__init__(self, sender, recipients, date, subject, unread, 
                           index
                           )
      self.__mb = mailbox
      self.__cached = None
      self.__deleted = 0
   
   def getMessage(self):
      """
         Returns the @spug.mail.message.Message instance associated with this
         message.
      """
      if not self.__cached:
         self.__cached = self.__mb.getMessage(self.index)
      return self.__cached

   def markAsRead(self):
      """
         Marks the message as read (if it is not alread marked as read).
      """
      if self.unread:
         self.unread = 0
         msg = self.getMessage()
         if msg.headers.has_key('status'):
            msg.headers['status'] = 'R' + msg.headers['status']
         else:
            msg.headers['status'] = 'R'
         self.__mb.writeMessage(self.index, msg)
         self.__mb.markIndexModified()

   def toggleDeleted(self):
      """
         Toggles the "deleted" flag, marking the message as deleted if it
         isn't, and unmarking it if it is.
      """
      self.__deleted = not self.__deleted
      self.__mb.markIndexModified()
   
   def isDeleted(self):
      """
         Returns true if the message is marked for deletion.
      """
      return self.__deleted

   def asMessageInfo(self):
      """
         Returns the item represented as the base class MessageInfo object.
      """
      return MessageInfo(self.sender, self.recipients, self.date, 
                         self.subject, self.unread, self.index
                         )

class Mailbox:
   """
      A mailbox manages the set of @spug.mail.message.Message objects
      in a shelve.
   """

   __intRx = re.compile('\d+')
   
   def __init__(self, dbFileName):
      self.fileName = dbFileName
      self.shelf = shelve.open(dbFileName)
      self.__index = None
      self.__indexModified = 0

   def __makeMailboxMessageInfo(self, item):
      return MailboxMessageInfo(item.sender,
                                item.recipients,
                                item.date,
                                item.subject,
                                item.unread,
                                item.index,
                                self
                                )
   
   def getIndex(self):
      """
         Returns the index, a list of @MailboxMessageInfo objects.
      """
      if not self.__index:
         try:
            index = self.shelf['index']
         except KeyError:
            index = []
         
         # convert all of the items to MailboxMessageInfo instances.
         for i in range(len(index)):
            item = index[i]
            index[i] = self.__makeMailboxMessageInfo(item)
         self.__index = index
                                       
      return self.__index

   def getMessage(self, index):
      """
         Gets the message at the given /index/ (an integer).
      """
      return self.shelf[`index`]                                       
         
   def writeMessage(self, index, message):
      """
         Writes the message at the given index.
      """
      self.shelf[`index`] = message

   def addMessage(self, message):
      """
         Adds a new message to the mailbox.  Returns the 
         @MailboxMessageInfo instance associated with it.
      """
      # first increment the index
      try:
         index = self.shelf['next']
      except KeyError:
         index = 0
      index = index + 1
      self.shelf['next'] = index
      
      # store the message
      self.writeMessage(index, message)

      # make sure that the message index has been loaded
      self.getIndex()
      
      # add the info to the main index
      info = self.__makeMailboxMessageInfo(message.asMessageInfo())
      info.index = index
      self.__index.append(info)
      self.__indexModified = 1
      
      return info

   def checkMailbox(self):
      """
         Checks the mailbox for messages that are in the database
         but not in the index, or in the index but not in the database.
         If any changes are made, writes the index.
         
         Returns a tuple of integers indicating the number of messages
         added and remove from the index.
      """
      
      # locate all items that are in the index but not in the database
      # and create a dictionary of valid indeces
      remove = []
      indexDict = {}
      index = self.getIndex()
      for item in index:
         if not self.shelf.has_key(`item.index`):
            remove.append(item)
         else:
            indexDict[item.index] = item
      
      # remove all non-existent entries from the index
      for item in remove:
         index.remove(item)
      
      # locate all items that are in the database but not in the index
      added = 0
      for key in self.shelf.keys():
         if self.__intRx.match(key):
            keyInt = int(key)
            if not indexDict.has_key(keyInt):
               msg = self.shelf[key]
               
               # get the "mailbox message info" for the item
               msgInfo = msg.asMessageInfo(keyInt)
               msgInfo = self.__makeMailboxMessageInfo(msgInfo)
               
               # add it to the index
               index.append(msgInfo)
               added = added + 1

      # mark the mailbox as modified if it was
      if added or remove:
         self.__indexModified = 1
               
      return added, len(remove)         

   def markIndexModified(self):
      self.__indexModified = 1

   def writeIndex(self):
      """
         Writes the entire index to the file if it has been modified.
      """
      if self.__indexModified:
         index = []
         for item in self.__index:
            if not item.isDeleted():
               index.append(item.asMessageInfo())
            else:
               try:
                  del self.shelf[`item.index`]
               except:
                  self.__handleDelError(item)
         self.shelf['index'] = index

   def __handleDelError(self, item):
      """
         Extremely safe way of handling errors that occur when deleting
         items from the shelf.  Rule 1 is that no matter what happens,
         we don't end up throwing an exception (which could block further
         activity).  Rule 2 is that whenever possible, we log the fact
         that there was an error.  Rule 3 is that we try to log the
         traceback.
      """
      try:
         err = open('spugmail.err', 'a')
         try:
            err.write('Error deleting message %d:\n' % item.index)
            import traceback, sys, string
            info = apply(traceback.format_exception, sys.exc_info())
            err.write(string.join(info, '\n') + '\n')
         finally:
            err.flush()
      except:
         pass

class PostOffice:
   """
      A collection of mailboxes.  Manages all of the file interactions for a 
      set of mailboxes.
   """
   
   specialMailboxes = ['Inbox', 'Sent', 'Drafts']
   
   def __init__(self, path):
      """
         Constructs a mailbox given the path to the mailbox directory.
      """
      self.__path = path
      self.__boxes = {}
   
   def __del__(self):
      for box in self.__boxes.values():
         box.writeIndex()
   
   def getMailboxNames(self):
      """
         Returns the list of strings detailing the names of all of the 
         mailboxes.
      """
      names = []
      for file in os.listdir(self.__path):
         names.append(file)
      return names
      
   def createMailbox(self, mailboxName):
      """
         Creates a new mailbox.
         
         parms:
            mailboxName::
               [string] name of the new mailbox.
      """
      mb = Mailbox(os.path.join(self.__path, mailboxName))
      self.__boxes[mailboxName] = mb
   
   def getMailbox(self, mailboxName):
      """
         Returns the named mailbox.  Raises a *KeyError* if the mailbox
         does not exist, unless the mailbox is a "special mailbox".  If it
         is, it is created.
         
         parms:
            mailboxName::
               [string] name of the mailbox to get
      """
      # check the cache
      if self.__boxes.has_key(mailboxName):
         return self.__boxes[mailboxName]
      
      # make sure that the mailbox exists on the disk
      fileName = os.path.join(self.__path, mailboxName)
      if not os.path.exists(fileName):
         
         # is it one of the special mailboxes?
         if mailboxName in self.specialMailboxes:
            
            # create it
            self.createMailbox(mailboxName)
            return self.__boxes[mailboxName]
         
         raise KeyError(mailboxName)
      
      # load the mailbox, store it and return it
      mb = Mailbox(fileName)
      self.__boxes[mailboxName] = mb
      return mb

   def hasMailbox(self, mailboxName):
      """
         Returns true if the mailbox exists.
         
         parms:
            mailboxName::
               [string] name of the mailbox in question
      """
      if self.__boxes.has_key(mailboxName):
         return 1
      
      if os.path.exists(mailboxName):
         return 1
      
      return 0
 